/*
   lex.c -- PIKT common text routines

   $Header: /var/cvs/PIKT/pikt/src/lib/text.c,v 1.10 2006/11/04 09:59:35 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "common.h"

/* strip backslash escape chars from the given scan command; this is to
   ensure consistency of treatment with escape processing done in begin,
   end, and rules sections, where backslash escapes are dealt with by
   the decompose() function
 */
char *
stripesc (char *input)
{
	static char output[MAXLINE+1];
	char c;
	register int i, j;
	register int escape;

	escape = FALSE;
	i = j = 0;
	while ((c = *(input+i))) {
		i++;
		if ((c == '\\') && ! escape) {
			escape = TRUE;
			continue;
		}
		output[j++] = c;
		escape = FALSE;
	}
	output[j] = '\0';
	return output;
}

/* for every backslash escape in a given string, \ -> \\
 */
char *
addesc (char *input)
{
	static char output[2*MAXLINE+1];
	char c;
	register int i, j;

	i = j = 0;
	while ((c = *(input+i))) {
		i++;
		if ((c == '\\')) {
			output[j++] = c;
		}
		output[j++] = c;
	}
	output[j] = '\0';
	return output;
}

/* for input str, replace all double quotes (") with DQTPROXY,
   and all single quotes (') with SQTPROXY, returning strlen
 */
int
dequote (signed char *str)
{
	register int i;
	register signed char c;

	if (! str) {
		i = ERROR;
	} else {
		i = 0;
		while ((c = *(str+i))) {
			if (c == '\"')
				*(str+i) = DQTPROXY;
			else if (c == '\'')
				*(str+i) = SQTPROXY;
			i++;
		}
	}
	return i;
}

/* for input str, replace all DQTPROXY with double quotes ("),
   and all SQTPROXY with single quotes ('), returning strlen
 */
int
requote (signed char *str)
{
	register int i;
	register signed char c;

	if (! str) {
		i = ERROR;
	} else {
		i = 0;
		while ((c = *(str+i))) {
			if (c == DQTPROXY)
				*(str+i) = '\"';
			else if (c == SQTPROXY)
				*(str+i) = '\'';
			i++;
		}
	}
	return i;
}
