/*
   macro.c -- PIKT piktc macro routines

   $Header: /var/cvs/PIKT/pikt/src/piktc/macro.c,v 1.20 2006/11/04 09:59:35 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktc.h"

/* in a macro definition, replace referenced macro parameters with their
   parametized versions; for example, with

   foo(A,B)	blah (A) blah (B)

   substitute '(A)' with 'fooMCRLNK1', since A is the foo macro's
   first parameter; and
   substitute '(B)' with 'fooMCRLNK2', since B is the foo macro's
   second parameter
 */
char *
parametize (char *mcrnam, unsigned int mcrnamlen, char *cmp, int linnum)
{
	static char fnnam[] = "parametize";
	char res[MAXBUF+1];
	char prm[MAXIDENT+1];
	register int icmp, imcr, iprm, ires;
	register char ccmp, cmcr, cprm;
	int prmlen;
	ident *idp;
	char *mcrnamprm;

	*prm = '\0';
	prmlen = 0;
	icmp = ires = iprm = 0;
	while ((ccmp = *(cmp+icmp++))) {
		if (ires >= MAXBUF) {
			if (linnum == NONE) {
				logexit(OUTPUT, LOG_ERR, ERRID_MACRO_PARAMETER_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "for host %s, for command-line macro =%s, parametized component too long", hostcurr, mcrnam);
			} else {
				logexit(OUTPUT, LOG_ERR, ERRID_MACRO_PARAMETER_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, for macro =%s, parametized component too long", cfgfilecurr, linnum+1, hostcurr, mcrnam);
			}
		}
		if (ccmp == '\(') {
			if (iprm) {
				*(res+ires++) = '\(';
				*(prm+iprm) = '\0';
				iprm = 0;
				while ((iprm < prmlen) && (cprm = *(prm+iprm++)))
					*(res+ires++) = cprm;
				*prm = '\0';
				prmlen = 0;
				iprm = 0;
			}
			if ((cprm = *(cmp+icmp)) && isalpha((int)cprm)) {
				*(prm+iprm++) = cprm;
				prmlen++;
				icmp++;
			} else {
				*(res+ires++) = ccmp;
			}
		} else if (ccmp == ')') {
			if (iprm) {
				*(prm+iprm) = '\0';
				iprm = 0;
			} else {
				*(res+ires++) = ccmp;
			}
			if (*(prm+iprm)) {
				mcrnamprm = mycalloc(mcrnamlen+1+prmlen+1, sizeof(char), fnnam);
				sprintf(mcrnamprm, "%s%c%s", mcrnam, MCRLNK, prm);
				if ((idp = lookupident(mcrnamprm, MCRCFG))) {
					*(res+ires++) = '=';
					imcr = 0;
					while ((cmcr = *(mcrnam+imcr++)))
						*(res+ires++) = cmcr;
					*(res+ires++) = MCRLNK;
					*(res+ires++) = *idp->definition->component;
				} else {
					*(res+ires++) = '\(';
					while ((iprm < prmlen) && (cprm = *(prm+iprm++)))
						*(res+ires++) = cprm;
					*(res+ires++) = ')';
				}
				free(mcrnamprm);
				*prm = '\0';
				prmlen = 0;
				iprm = 0;
			}
		} else if (isalnum((int)ccmp) || strchr("-_.",ccmp)) {
			if (iprm) {
				*(prm+iprm++) = ccmp;
				prmlen++;
			} else {
				*(res+ires++) = ccmp;
			}
		} else {
			if (iprm) {
				*(res+ires++) = '\(';
				*(prm+iprm) = '\0';
				iprm = 0;
				while ((iprm < prmlen) && (cprm = *(prm+iprm++)))
					*(res+ires++) = cprm;
				*prm = '\0';
				prmlen = 0;
				iprm = 0;
			}
			*(res+ires++) = ccmp;
		}
	}
	*(res+ires) = '\0';
	return mystrdup(res, fnnam);
}

/* do all macro expansions on the given line; break the line into up to three
   parts: head, macro (with arguments, if any), and tail; concatenate
   together the head, the macro-expanded macro part, and the expanded tail;
   return the expanded line
 */
char *
expand (char *line, int linnum)
{
	static char fnnam[] = "expand";
	char head[MAXBUF+1], mcrnam[MAXBUF+1], mcrarg[MAXBUF+1], tail[MAXBUF+1];
	int iline, ihead, imcrnam, imcrarg, itail;
	char cline;
	int lp, rp, dqt;
	int escape;
	char *expline, *expbody, *exptail;

	escape = FALSE;
	lp = rp = dqt = 0;
	iline = ihead = imcrnam = imcrarg = itail = 0;
	head[0] = mcrnam[0] = mcrarg[0] = tail[0] = '\0';
	while ((cline = *(line+iline++))) {
		if (iline >= MAXBUF) {
			logexit(OUTPUT, LOG_ERR, ERRID_MACRO_EXPANSION_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "for host %s, macro expansion too long: %s", hostcurr, line);
		}
		if (cline == '\\') {
			if (! escape) {
				escape = TRUE;
				if (    *(line+iline)
				     && (*(line+iline) == '=')
				     && ! tail[0]
				   ) {
					continue;
				}
			} else {
				escape = FALSE;
			}
		} else if (cline == '=') {
			if (! escape) {
				if (mcrnam[0]) {
					if (! mcrarg[0] && ! tail[0])
						tail[0] = TRUE;
				} else {
					/* is macro if next char is alpha or '_' */
					if (    *(line+iline)
					     && (    isalpha((int)*(line+iline))
					          || (*(line+iline) == '_')
					        )
					     && ( ! mcrarg[0] && ! tail[0] )
					   ) {
						mcrnam[0] = TRUE;
						continue;
					}
				}
			}
			escape = FALSE;
		} else if (cline == '\(') {
			if (! escape && mcrnam[0]) {
				if (even(dqt))
					lp++;
				if (mcrnam[0] && ! mcrarg[0] && ! tail[0]) {
					mcrarg[0] = TRUE;
					*(mcrarg+imcrarg++) = cline;
					continue;
				}
			}
			escape = FALSE;
		} else if (cline == ')') {
			if (! escape && mcrarg[0]) {
				if (even(dqt))
					rp++;
				if (lp <= rp) {
					if (mcrarg[0] && ! tail[0]) {
						tail[0] = TRUE;
						*(mcrarg+imcrarg++) = cline;
						continue;
					}
				}
			}
			escape = FALSE;
		} else if (cline == '"') {
			if (! escape && mcrarg[0]) {
				dqt++;
			}
			escape = FALSE;
		} else {
			escape = FALSE;
		}
		if (tail[0]) {
			*(tail+itail++) = cline;
		} else if (mcrarg[0]) {
			*(mcrarg+imcrarg++) = cline;
		} else if (mcrnam[0]) {
			if (    isalnum((int)cline)
			     || strchr("-_.", cline)
			     || (cline == MCRLNK)
			     || (    (cline <= 32)
			          && (*(line+iline-2) == MCRLNK)
			          && (*(line+iline-3) != MCRLNK)
			        )
			   ) {
				*(mcrnam+imcrnam++) = cline;
				/* the special =_ macro */
				if ((cline == '_') && (imcrnam == 1)) {
					tail[0] = TRUE;
				}
				/* a parameter macro */
				if (*(line+iline-2) == MCRLNK) {
					tail[0] = TRUE;
				}
			} else {
				*(tail+itail++) = cline;
			}
		} else {
			*(head+ihead++) = cline;
		}
	}
	/* terminate the head, mcrnam, mcrarg & tail strings */
	head[ihead] = mcrnam[imcrnam] = mcrarg[imcrarg] = tail[itail] = '\0';
	if (mcrnam[0])
		expbody = mcrexpand(mcrnam, imcrnam, mcrarg, imcrarg, linnum);
	else
		expbody = mystrdup(nullstring, fnnam);
	if (tail[0])
		exptail = expand(tail, linnum);
	else
		exptail = mystrdup(nullstring, fnnam);
	expline = mycalloc(strlen(head)+strlen(expbody)+strlen(exptail)+1, sizeof(char), fnnam);
	sprintf(expline, "%s%s%s", head, expbody, exptail);
	free(expbody);
	free(exptail);
	return expline;
}

/* for the given macro and macro arguments, if any, return its expansion
 */
char *
mcrexpand (char *mcrnam, int mcrnamlen, char *mcrarg, int mcrarglen, int linnum)
{
	static char fnnam[] = "mcrexpand";
	ident *idp;
	def *dfp;
	char *mcrexp;
	char *compexp;
	char *res;
	int mcrexplen;
	int complen;

	if (! *mcrnam) {
		res = mystrdup(nullstring, fnnam);
	} else if (    ( cfgcurr == MCRCFG )
	            && (    ! strcmp(mcrnam, "piktnullchar")
	                 || ! strcmp(mcrnam, "piktspacechar")
	                 || ! strcmp(mcrnam, "pikttabchar")
	                 || ! strcmp(mcrnam, "piktnewlinechar")
	                 || ! strcmp(mcrnam, "_")
	               )
	          ) {
		res = mycalloc((size_t)(mcrnamlen+mcrarglen+2), sizeof(char), fnnam);
		sprintf(res, "=%s%s", mcrnam, mcrarg);
	} else if (! (idp = lookupident(mcrnam, MCRCFG))) {
		res = (char *) NULL;	/* to squelch compiler warning "'res'
					   might be unused in this function"
					 */
		if (linnum == NONE) {
			logexit(OUTPUT, LOG_ERR, ERRID_MACRO_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "for host %s, unknown macro =%s", hostcurr, mcrnam);
		} else {
			logexit(OUTPUT, LOG_ERR, ERRID_MACRO_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, unknown macro =%s", cfgfilecurr, linnum+1, hostcurr, mcrnam);
		}
	} else if (idp->type != MCR) {
		res = mycalloc((size_t)(mcrnamlen+mcrarglen+2), sizeof(char), fnnam);
		sprintf(res, "=%s%s", mcrnam, mcrarg);
	} else {
		if (idp->called) {
			if (linnum == NONE) {
				logexit(OUTPUT, LOG_ERR, ERRID_MACRO_SELF_REFERENCING, __FILE__, __LINE__, fnnam, FALSE, "for host %s, for macro =%s, illegal self reference", hostcurr, mcrnam);
			} else {
				logexit(OUTPUT, LOG_ERR, ERRID_MACRO_SELF_REFERENCING, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, for macro =%s, illegal self reference", cfgfilecurr, linnum+1, hostcurr, mcrnam);
			}
		} else {
			idp->called = TRUE;	/* entering this macro call */
		}
		setmcrargs(mcrnam, mcrnamlen, idp->prmnum, mcrarg, linnum);
		mcrexplen = 0;
		mcrexp = mycalloc(MAXBUF+1, sizeof(char), fnnam);
		foreach (dfp, idp->definition) {
			if (*(dfp->component))
				compexp = expand(dfp->component, linnum);
			else
				compexp = mystrdup(nullstring, fnnam);
			if ((complen = strlen(compexp))) {
				if (mcrexplen + complen + 2 >= MAXBUF) {
					if (linnum == NONE) {
						logexit(OUTPUT, LOG_ERR, ERRID_MACRO_EXPANSION_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "for host %s, for command-line macro =%s, macro expansion too long", hostcurr, mcrnam);
					} else {
						logexit(OUTPUT, LOG_ERR, ERRID_MACRO_EXPANSION_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, for macro =%s, macro expansion too long", cfgfilecurr, linnum+1, hostcurr, mcrnam);
					}
				}
				strcat(mcrexp, compexp);
				strcat(mcrexp, " ");
				mcrexplen += complen+1;
			}
			free(compexp);
		}
		if (mcrexplen) {
			*(mcrexp+mcrexplen-1) = '\0';	/* kill the trailing space */
			res = mystrdup(mcrexp, fnnam);
		} else {
			res = mystrdup(nullstring, fnnam);
		}
		free(mcrexp);
		unsetmcrargs(mcrnam, mcrnamlen, idp->prmnum, linnum);
		idp->called = FALSE;		/* leaving this macro call */
	}
	return res;
}

/* parse the given macro arguments, if any; assign each arg to the
   appropriate macro parameter ident
 */
void
setmcrargs (char *mcrnam, int mcrnamlen, int prmsnum, char *args, int linnum)
{
	static char fnnam[] = "setmcrargs";
	int prmnum;
	char arg[MAXMCRBUF+1];
	register int iargs, iarg;
	register char cargs, carg;
	char *mcrnamprmnum;
	ident *idp;
	int lp, rp, dqt;
	int escape;

	if (prmsnum == 0)
		return;
	escape = FALSE;
	lp = rp = dqt = 0;
	prmnum = 0;
	iarg = 0;
	iargs = 0;
	if (*args == '\(') {
		iargs = 1;	/* skip leading lp */
	} else {
		if (linnum == NONE) {
			logexit(OUTPUT, LOG_ERR, ERRID_MACRO_ARG_INVALID, __FILE__, __LINE__, fnnam, FALSE, "for host %s, for command-line macro =%s, badly formed macro args (no opening parenthesis)", hostcurr, mcrnam);
		} else {
			logexit(OUTPUT, LOG_ERR, ERRID_MACRO_ARG_INVALID, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, for macro =%s, badly formed macro args (no opening parenthesis)", cfgfilecurr, linnum+1, hostcurr, mcrnam);
		}
	}
	while ((cargs = *(args+iargs++))) {
		if (iarg >= MAXMCRBUF) {
			if (linnum == NONE) {
				logexit(OUTPUT, LOG_ERR, ERRID_MACRO_ARG_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "for host %s, for command-line macro =%s, macro argument too long", hostcurr, mcrnam);
			} else {
				logexit(OUTPUT, LOG_ERR, ERRID_MACRO_ARG_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, for macro =%s, macro argument too long", cfgfilecurr, linnum+1, hostcurr, mcrnam);
			}
		}
		if (cargs == '\\') {
			if (! escape) {
				escape = TRUE;
			}
			else {
				escape = FALSE;
				*(arg+iarg++) = cargs;
			}
		} else if (cargs == '\(') {
			if (! escape) {
				if (even(dqt))
					lp++;
			}
			*(arg+iarg++) = cargs;
			escape = FALSE;
		} else if ((cargs == ',') || (cargs == ')')) {
			if (! escape) {
				if (even(dqt)) {
					if (cargs == ')')
						rp++;
					if (    (lp <= rp)
					     && (    (    (cargs == ',')
					               && (prmsnum > 1)
					             )
					          || ! *(args+iargs)
						)
					   ) {
						if (++prmnum > prmsnum) {
							if (linnum == NONE) {
								logexit(OUTPUT, LOG_ERR, ERRID_MACRO_ARG_COUNT_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "for host %s, for command-line macro =%s, too many macro arguments", hostcurr, mcrnam);
							} else {
								logexit(OUTPUT, LOG_ERR, ERRID_MACRO_ARG_COUNT_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, for macro =%s, too many macro arguments", cfgfilecurr, linnum+1, hostcurr, mcrnam);
							}
						}
						mcrnamprmnum = mycalloc((size_t)(mcrnamlen+3), sizeof(char), fnnam);
						sprintf(mcrnamprmnum, "%s%c%c", mcrnam, MCRLNK, prmnum);
						if (! (idp = lookupident(mcrnamprmnum, MCRCFG))) {
							if (linnum == NONE) {
								logexit(OUTPUT, LOG_ERR, ERRID_MACRO_PARAMETER_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "for host %s, for command-line macro =%s, parameter %d not registered", hostcurr, mcrnam, prmnum);
							} else {
								logexit(OUTPUT, LOG_ERR, ERRID_MACRO_PARAMETER_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, for macro =%s, parameter %d not registered", cfgfilecurr, linnum+1, hostcurr, mcrnam, prmnum);
							}
						} else {
							*(arg+iarg) = '\0';
							iarg--;
							while (   ((carg = *(arg+iarg)) == ' ')
							       || (carg == '\t')) {
								*(arg+iarg) = '\0';
								iarg--;
							}
							regcomponent(idp, mystrdup(arg, fnnam));
						}
						iarg = 0;
						free(mcrnamprmnum);
					} else {
						*(arg+iarg++) = cargs;
					}
				} else {
					*(arg+iarg++) = cargs;
				}
			} else {
				*(arg+iarg++) = cargs;
			}
			escape = FALSE;
		} else if ((cargs == ' ') || (cargs == '\t')) {
			if (odd(dqt) || (iarg != 0))
				*(arg+iarg++) = cargs;
			escape = FALSE;
		} else if (cargs == '"') {
			if (! escape)
				dqt++;
			*(arg+iarg++) = cargs;
			escape = FALSE;
		} else {
			*(arg+iarg++) = cargs;
			escape = FALSE;
		}
	}
	if (*(args+iargs-2) != ')') {
		if (linnum == NONE) {
			logexit(OUTPUT, LOG_ERR, ERRID_MACRO_ARG_INVALID, __FILE__, __LINE__, fnnam, FALSE, "for host %s, for command-line macro =%s, badly formed macro args (no closing parenthesis)", hostcurr, mcrnam);
		} else {
			logexit(OUTPUT, LOG_ERR, ERRID_MACRO_ARG_INVALID, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, for macro =%s, badly formed macro args (no closing parenthesis)", cfgfilecurr, linnum+1, hostcurr, mcrnam);
		}
	}
	if (prmnum < prmsnum) {
		if (linnum == NONE) {
			logexit(OUTPUT, LOG_ERR, ERRID_MACRO_ARG_COUNT_TOO_SMALL, __FILE__, __LINE__, fnnam, FALSE, "for host %s, for command-line macro =%s, too few macro arguments", hostcurr, mcrnam);
		} else {
			logexit(OUTPUT, LOG_ERR, ERRID_MACRO_ARG_COUNT_TOO_SMALL, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, for macro =%s, too few macro arguments", cfgfilecurr, linnum+1, hostcurr, mcrnam);
		}
	}
	return;
}

/* after a macro call, clear all macro parameter idents
 */
void
unsetmcrargs (char *mcrnam, int mcrnamlen, int prmsnum, int linnum)
{
	static char fnnam[] = "unsetmcrargs";
	int prmnum;
	char *mcrnamprmnum;
	ident *idp;
	def *dfp, *tdfp;

	for (prmnum=1; prmnum<=prmsnum; prmnum++) {
		mcrnamprmnum = mycalloc((size_t)(mcrnamlen+3), sizeof(char), fnnam);
		sprintf(mcrnamprmnum, "%s%c%c", mcrnam, MCRLNK, prmnum);
		if (! (idp = lookupident(mcrnamprmnum, MCRCFG))) {
			if (linnum == NONE) {
				logexit(OUTPUT, LOG_ERR, ERRID_MACRO_PARAMETER_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "for host %s, for command-line macro =%s, parameter %d not registered", hostcurr, mcrnam, prmnum);
			} else {
				logexit(OUTPUT, LOG_ERR, ERRID_MACRO_PARAMETER_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d, for host %s, for macro =%s, parameter %d not registered", cfgfilecurr, linnum+1, hostcurr, mcrnam, prmnum);
			}
		} else {
			dfp = idp->definition;
			while (dfp) {
				tdfp = dfp;
				free(dfp->component);
				dfp = dfp->next;
				free(tdfp);
			}
			idp->definition = (def *) NULL;
			idp->defcurr = (def *) NULL;
		}
		free(mcrnamprmnum);
	}
	return;
}
