%{

/*
   alt.l -- PIKT alarm script lexer

   $Header: /var/cvs/PIKT/pikt/src/piktd/alt.l,v 1.64 2006/11/04 09:59:36 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   ----------------------------------------------------------------------------

   this is the sausage factory; not for the squeamish; enter at your own risk!

   on our test irix system, flex failed to match regexps of the form
   '{D}{x,y}', which led to parser failures; hence substitutions of the
   form '{D}+' below

 */

#include "../lib/common.h"
#include "piktd.h"
#include "alt.tab.h"

char _altlex_[] = "altlex";

#undef YY_INPUT
#define YY_INPUT(buf, result, max_size) { \
		signed char c; \
		if (command) { \
			c = altinput(); \
		} else { \
			c = fgetc(altin); \
		} \
		result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
	}

#define RETURN(token) altlval.sval = mystrdup(yytext, _altlex_); return token

/* internally, every Pikt value is considered an array element; in the case
   of scalars, we pretend as if $foo were really written as $foo[0]
 */
#define FAKEARRAY unput(']'); unput('0'); unput('['); unput('['); unput('[')

int ii, jj, kk;
int d;
int v;
char cc;
char *s1, *s2;
char *yyt;
int yyl;

char outstr[MAXBUF+1];

char *nam;

int fn = 0;

int isarray;
int isfakearray;

void alterror(const char *);
signed int altinput(void);

%}

W		[ \t]
nW		[^ \t]
WN		[ \t\n]
nWN		[^ \t\n]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]
DLU		[0-9A-Za-z_]
nDLU		[^0-9A-Za-z_]
nB		[^\[\]]

%x CS DS SS FS KFS CMNT2

%option noyywrap

%%

<*>^{L}[-_\.0-9A-Za-z]*		{ if (isscript) {
					yyless(0);
				  } else {
				  	BEGIN(INITIAL);
				  	RETURN(_STANZA);
				  }
				}

<INITIAL,DS,SS>status		{ return _STATUS; }
<INITIAL,DS,SS>active		{ return _ACTIVE; }
<INITIAL,DS,SS>inactive		{ return _INACTIVE; }

<INITIAL,DS,SS>level		{ return _LEVEL; }
<INITIAL,DS,SS>emerg|emergency	{ return _EMERG; }
<INITIAL,DS,SS>alert|urgent	{ return _ALERT; }
<INITIAL,DS,SS>crit|critical	{ return _CRIT; }
<INITIAL,DS,SS>err|error	{ return _ERR; }
<INITIAL,DS,SS>warn|warning	{ return _WARNING; }
<INITIAL,DS,SS>notice		{ return _NOTICE; }
<INITIAL,DS,SS>info		{ return _INFO; }
<INITIAL,DS,SS>debug		{ return _DEBUG; }

<INITIAL,DS,SS>task		{ return _TASK; }

<INITIAL,DS,SS>input		{ return _INPUT; }
<INITIAL,DS,SS>proc|process	{ BEGIN(CS);
				  return _PROCESS;
				}
<INITIAL,DS,SS>file		{ return _REGFILE; }
<INITIAL,DS,SS>log|logfile	{ return _LOGFILE; }

<INITIAL,DS,SS>filter		{ BEGIN(CS);
				  return _FILTER;
				}

<INITIAL,DS,SS>seps|separators	{ return _SEPARATORS; }
<INITIAL,DS,SS>dat|data		{ BEGIN(DS);
				  return _DAT;
				}
<INITIAL,DS,SS>keys		{ BEGIN(DS);
				  return _KEYS;
				}

<INITIAL,DS,SS>init		{ return _INIT; }
<INITIAL,DS,SS,CS>begin		{ BEGIN(INITIAL);
				  return _BEGIN;
				}
<INITIAL,DS,SS,CS>end		{ BEGIN(INITIAL);
				  return _END;
				}
<INITIAL,DS,SS,CS>rule		{ BEGIN(INITIAL);
				  return _RULE;
				}

<INITIAL,DS,SS>do		{ return _DO; }
<INITIAL,DS,SS>if		{ return _IF; }
<INITIAL,DS,SS>if{WN}*"("	{ unput('[');
				  return _IF;
				}
<INITIAL,DS,SS>(elif|elsif|elseif) {
				  return _ELIF;
				}
<INITIAL,DS,SS>(elif|elsif|elseif){WN}*"(" {
				  unput('[');
				  return _ELIF;
				}
<INITIAL,DS,SS>else		{ return _ELSE;	}
<INITIAL,DS,SS>endif|fi		{ return _ENDIF; }
<INITIAL,DS,SS>while		{ return _WHILE; }
<INITIAL,DS,SS>while{WN}*"("	{ unput('[');
				  return _WHILE;
				}
<INITIAL,DS,SS>endwhile		{ return _ENDWHILE; }
<INITIAL,DS,SS>repeat		{ return _REPEAT; }
<INITIAL,DS,SS>until		{ return _UNTIL; }
<INITIAL,DS,SS>until{WN}*"("	{ unput('[');
				  return _UNTIL;
				}
<INITIAL,DS,SS>for		{ return _FOR; }
<INITIAL,DS,SS>for{WN}*"("	{ return _FORLP; }
<INITIAL,DS,SS>foreach		{ return _FOREACH; }
<INITIAL,DS,SS>endfor|endforeach { return _ENDFOR; }
<INITIAL,DS,SS>break		{ return _BREAK; }
<INITIAL,DS,SS>continue|cont	{ return _CONTINUE; }
<INITIAL,DS,SS>switch		{ return _SWITCH; }
<INITIAL,DS,SS>switch{WN}*"("	{ return _SWITCHLP; }
<INITIAL,DS,SS>case		{ return _CASE;	}
<INITIAL,DS,SS>default		{ return _DEFAULT; }
<INITIAL,DS,SS>breakswitch|breaksw { return _BREAKSWITCH; }
<INITIAL,DS,SS>endswitch|endsw	{ return _ENDSWITCH; }
<INITIAL,DS,SS>again		{ return _AGAIN; }
<INITIAL,DS,SS>leave		{ return _LEAVE; }
<INITIAL,DS,SS>redo		{ return _REDO; }
<INITIAL,DS,SS>next		{ return _NEXT;	}
<INITIAL,DS,SS>skip		{ return _SKIP;	}
<INITIAL,DS,SS>last		{ return _LAST;	}
<INITIAL,DS,SS>pause		{ return _PAUSE; }
<INITIAL,DS,SS>quit		{ return _QUIT;	}
<INITIAL,DS,SS>die		{ return _DIE; }

<INITIAL,DS,SS>output		{ return _OUTPUT; }
<INITIAL,DS,SS>mail		{ return _MAIL; }
<INITIAL,DS,SS>print		{ return _PRINT; }
<INITIAL,DS,SS>piktlog		{ return _PIKTLOG; }
<INITIAL,DS,SS>syslog		{ return _SYSLOG; }

<INITIAL,DS,SS>set/{WN}+(#|$){L}{DLU}*{WN}+"=" {
				  isfakearray = TRUE;
				  BEGIN(SS);
				  return _SET;
				}
<INITIAL,DS,SS>set		{ isfakearray = FALSE;
				  BEGIN(SS);
				  return _SET;
				}
<INITIAL,DS,SS>unset|delete	{ BEGIN(SS);
				  return _UNSET;
				}

<INITIAL,DS,SS>exec		{ BEGIN(INITIAL);
				  return _EXEC;
				}
<INITIAL>wait			{ return _WAIT; }

<INITIAL,DS,SS>for{WN}+${L}{DLU}*{WN}+in{WN}+#(members|keys|skeys|sortkeys|rkeys|reversekeys){WN}*"(" {
				  /* for $k in #(members|...)( -->
				     foreach #(members|...)($k,
				   */
				  /* first, capture #(members|...) in s2 */
				  s2 = strchr(yytext, '#');
				  kk = 1;
				  while (isalpha((int)*(s2+kk))) {
					kk++;
				  }
				  *(s2+kk) = '\0';
				  /* next, capture $foo in s1 */
				  s1 = strchr(yytext, '$');
				  jj = 0;
				  while (! isspace((int)*(s1+jj))) {
					jj++;
				  }
				  *(s1+jj) = '\0';
				  /* write the replacement foreach string */
				  /* 11, because strlen("foreach ") plus
				     '(' plus ',' plus null char
				   */
				  yyt = (char *) mymalloc((unsigned)(11+jj+kk), _altlex_);
				  sprintf(yyt, "foreach %s(%s,", s2, s1);
				  /* push it back to the lexer input stream */
				  ii = 10+jj+kk;
				  while (ii) {
					unput(*(yyt+ii-1));
					ii--;
				  }
				  free(yyt);
				}

<CS,DS>\"[^\n]+\"/{WN}*		{ yytext[yyleng-1] = '\0';
				  BEGIN(INITIAL);
				  altlval.sval = mystrdup(yytext+1, _altlex_);
				  return _STR;
				}

<DS>"["+{WN}*\"[^\"\n]+\"/{WN}*"]" {
				  ii = 1;
				  while ((*(yytext+ii) == '\[') || isspace((int)*(yytext+ii)))
					ii++;
				  d = decompose(yytext+ii, outstr);
				  for (ii=d; ii>=0; ii--)
					unput(*(outstr+ii));
				  return '[';
				}
<DS>({D}+|$|$-{D}+)/[^\]0-9A-Za-z_] {
				  /* <DS>({D}{1,5}|$|$-{D}{1,5})/[^\]0-9A-Za-z_] */
				  BEGIN(INITIAL);
				  RETURN(_POSITION);
				}
<DS>"["({D}+|$|$-{D}+)"]" 	{ /* <DS>"["({D}{1,5}|$|$-{D}{1,5})"]" */
				  BEGIN(INITIAL);
				  RETURN(_POSITION);
				}
<DS>{D}+","({D}+|$)		{ BEGIN(INITIAL);
				  RETURN(_RANGE);
				}
<DS>"["{D}+","({D}+|$)"]"	{ BEGIN(INITIAL);
				  RETURN(_RANGE);
				}

<DS>(#|$){L}{DLU}*"[" 		{ yytext[yyleng-1] = '\0';
				  nam = (char *) mymalloc((yyleng-2+3) * sizeof(char), _altlex_);
				  sprintf(nam, " %s ", yytext+1);
				  if (strstr(" inlin inline inputline outlin outline outputline rdlin rdline readline ", nam)) {
					alterror("syntax error");
				  }
				  free(nam);
				  altlval.sval = mystrdup(yytext, _altlex_);
				  unput('[');
				  unput('[');
				  unput('[');
				  return (*altlval.sval == '#' ? _VARNUM : _VARSTR);
				}

<INITIAL,SS>"="			{ BEGIN(INITIAL);
				  return yytext[0];
				}

<INITIAL,SS>(#|$){L}{DLU}*({WN}*|"["[^\n]+"]"{WN}*)("+="|"-="|"*="|"/="|"%="|"\*\*="|"<<="|">>="|"&="|"|="|"^="|".=") {
				  yyt = mystrdup(yytext, _altlex_);
				  yyl = yyleng;
				  ii = yyl - 2;
				  while (ii >= 0)
					unput(*(yyt+ii--));
				  unput(' ');
				  unput('=');
				  unput(' ');
				  cc = *(yyt+yyl-3);
				  ii = yyl - ((isspace((int)cc) || isalnum((int)cc)) ? 3 : 4);
				  while (ii >= 0)
					unput(*(yyt+ii--));
				  free(yyt);
				}

<INITIAL,DS,SS,FS>\'[^\'\n]*\'	{ yytext[yyleng-1] = '\0';
				  altlval.sval = mystrdup(yytext+1, _altlex_);
				  return _STR;
				}

<FS>[A-Z][0-9A-Z]*		{ nam = (char *) mymalloc((yyleng+2) * sizeof(char), _altlex_);
				  sprintf(nam, "?%s", yytext);
				  altlval.sval = mystrdup(nam, _altlex_);
				  free(nam);
				  return _FLHNDL;
				}

<INITIAL,SS,FS>${D}+		{ /* <INITIAL,SS,FS>${D}{1,5} */
				  altlval.sval = mystrdup("$", _altlex_);
				  yyt = mystrdup(yytext, _altlex_);
				  ii = yyleng-1;
				  unput(']');
				  while (ii > 0)
					unput(*(yyt+ii--));
				  free(yyt);
				  unput('[');
				  unput('[');
				  unput('[');
				  return _VALSTR;
				}

<INITIAL,DS,SS,FS>(#|$){L}{DLU}*/{WN}*"=" {
				  altlval.sval = mystrdup(yytext, _altlex_);
				  FAKEARRAY;
				  return (*altlval.sval == '#' ? _VARNUM : _VARSTR);
				}

<INITIAL,DS,SS,FS>(#|$){L}{DLU}* {
				  if ((YY_START == INITIAL) && (yytext[0] == '#') && (vardattype(yytext+1, yyleng-1, TRUE) != NUMBER)) {
					BEGIN(CMNT2);
				  } else {
					  altlval.sval = mystrdup(yytext, _altlex_);
					  FAKEARRAY;
					  return (*altlval.sval == '#' ? _VARNUM : _VARSTR);
				  }
				}

<INITIAL,SS,FS>$"["		{ altlval.sval = mystrdup("$", _altlex_);
				  unput('[');
				  unput('[');
				  unput('[');
				  return _VALSTR;
				}

<INITIAL,SS,FS>(#|$){L}{DLU}*"[" {
				  yytext[yyleng-1] = '\0';
				  if ((YY_START == INITIAL) && (yytext[0] == '#') && (vardattype(yytext+1, yyleng-2, TRUE) != NUMBER)) {
					BEGIN(CMNT2);
				  } else {
					nam = (char *) mymalloc((yyleng-2+3) * sizeof(char), _altlex_);
					sprintf(nam, " %s ", yytext+1);
					if (strstr(" inlin inline inputline outlin outline outputline rdlin rdline readline ", nam)) {
						alterror("syntax error");
					}
					free(nam);
					altlval.sval = mystrdup(yytext, _altlex_);
					unput('[');
					unput('[');
					unput('[');
					return (*altlval.sval == '#' ? _VARNUM : _VARSTR);
				  }
				}

<INITIAL,FS>(%|@){L}{DLU}*{nDLU} {
				  if (yytext[yyleng-1] == '\[')
					isarray = TRUE;
				  else
				  	isarray = FALSE;
				  altlval.sval = mystrdup(yytext, _altlex_);
				  unput(yytext[yyleng-1]);
				  *(altlval.sval+yyleng-1) = '\0';
				  if (! isarray) {
					FAKEARRAY;
				  }
				  switch (vardattype(altlval.sval+1, yyleng-2, FALSE)) {
					case NUMBER:
						return (*altlval.sval == '%' ? _VARHSTNUM : _VARPRCNUM);
						break;
				  	case STRING:
						return (*altlval.sval == '%' ? _VARHSTSTR : _VARPRCSTR);
						break;
					case ERROR:
						free(altlval.sval);
						altlval.sval = mystrdup(err, _altlex_);
						return _STR;
						break;
				  }
				}

<INITIAL,DS,SS,FS>"["/{WN}*($|\') { return yytext[0]; }
<INITIAL,DS,SS,FS>"["		{ return yytext[0]; }
<INITIAL,DS,SS,FS>"[[["		{ return '['; }
<INITIAL,DS,SS,FS>{D}+		{ /* <INITIAL,DS,SS,FS>{D}{1,5}	*/
				  RETURN(_INT);
				}
<INITIAL,DS,SS,FS>"]"		{ return yytext[0]; }

<INITIAL,SS,FS>("+"|"-"|"~")	{ return yytext[0]; }

<INITIAL,SS,FS>("*"|"/"|"%"|"**"|"&"|"|"|"^"|"<<"|">>") {
				  RETURN(_NOP);
				}

<INITIAL,SS,FS>"."		{ RETURN(_SOP);	}

<INITIAL,SS,FS>(-e|-z|-f|-d|-c|-b|-p|-l|-S) {
				  RETURN(_FOP);
				}

<INITIAL,SS,FS>("=="|">"|"<"|">="|"<="|"!>"|"!>="|"!<"|"!<="|"!="|"<>") {
				  RETURN(_NREL);
				}
<INITIAL,SS,FS>("=~"|"=~~"|"!~"|"!~~"|eq|ne|lt|gt|le|ge) {
				  RETURN(_SREL);
				}

<INITIAL,SS,FS>(({D}+)|({D}*\.{D}+)|((({D}+)|({D}*\.{D}+))\/(({D}+)|({D}*\.{D}+))))%* {
				  RETURN(_NUM);
				}

<INITIAL,SS,FS>#(keys|skeys|sortkeys|rkeys|reversekeys){WN}*"("{WN}*${L}{DLU}*{WN}*,{WN}*(#|$|%|@){L}{DLU}* {
				  /* set KEYSFN() or SORTKEYSFN() for the
				     assoc array var (the second var) in
				     #keys()|#skeys()|..., then push all
				     matched text back onto the lexer
				     input stream
				   */
				  s1 = strchr(yytext, ',');
				  if (! (s2 = strchr(s1, '#'))) {
					if (! (s2 = strchr(s1, '$'))) {
						if (! (s2 = strchr(s1, '%')))
							s2 = strchr(s1, '@');
					}
				  }
				  if ((v = varndx(s2+1)) > 0) {
					if (yytext[1] == 'k')	/* #keys */
						KEYSFN(v) = TRUE;
					else			/* #skeys|... */
						SORTKEYSFN(v) = TRUE;
				  } else {
					logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_UNKNOWN, __FILE__, __LINE__, _altlex_, FALSE, "unknown or illegal var %s in keys() function", s2+1);
					STATUS = INACTIVE;
				  }
				  BEGIN(KFS);
				  yyless(0);
				}

<INITIAL,SS,FS,KFS>(#|$){L}{DLU}*/"(" {
				  yyt = (char *) mycalloc((unsigned)(yyleng-1+3), sizeof(char), _altlex_);
				  sprintf(yyt, " %s ", yytext+1);
				  if ((yytext[0] == '#') && ! strstr(fnnumnames, yyt)) {
					
					free(yyt);
					BEGIN(CMNT2);
				  } else {
					free(yyt);
					BEGIN(FS);
					RETURN(yytext[0] == '#' ? _FNNUM : _FNSTR);
				  }
				}
<SS,FS>"("			{ fn++;
				  return '(';
				}
<SS,FS>")"			{ fn--;
				  if (! fn)
					BEGIN(INITIAL);
				  return ')';
				}

<INITIAL,SS,FS>\"[^\n\"]*\"	{ if ((d = decompose(yytext, outstr)) > 0) {
					for (ii=d; ii>=0; ii--)
						unput(*(outstr+ii));
				  } else {
					*(outstr+strlen(outstr)-1) = '\0';
					altlval.sval = mystrdup(outstr+1, _altlex_);
					return _STRVAR;
				  }
				}

<INITIAL>"("{WN}*{nWN}+{WN}*(=~|=~~|!~|!~~|eq|ne|lt|gt|le|ge){WN}*{nWN}+{WN}*")"{WN}* { 
				  yyless(1);
				  return '[';
				}

<INITIAL>"("([-+*/%$#@~^><)(,.0-9A-Za-z_\n\t\[\] ]|![^=~><]|"&"[^&]|"|"[^|])+ {
				  if (condparen(yytext)) {
					yyless(1);
					return '[';
				  } else {
					yyless(1);
					return yytext[0];
				  }
				}

<INITIAL>")"			{ return yytext[0]; }

<INITIAL,SS,FS>"!"		{ return _NOT; }
<INITIAL,SS,FS>"&&"		{ return _AND; }
<INITIAL,SS,FS>"||"		{ return _OR; }

<INITIAL,CS,DS,SS,FS,KFS>#[^A-Za-z\n].*\n {
				  lineno++;
				}
<INITIAL,CS,DS,SS,FS,KFS>#\n	{ lineno++; }

<CMNT2>.*\n			{ lineno++;
				  BEGIN(INITIAL);
				}

<*>","				{ return yytext[0]; }

<*>{W}				{ ; }
<*>\n				{ lineno++; }

<*>.				{ return yytext[0]; }

%%

/* alt error handler
 */
void
alterror (const char *message)
{
	static char fnnam[] = "alterror";
	static int l = 0;
	static char *errtext1;
	static char *errtext2;

	if (l != lineno)
		l = lineno;
	errtext1 = mystrdup(altlval.sval, fnnam);
	if (! strncmp(yytext, "\[\[\[0]", 5)) {
		errtext2 = mystrdup(nullstring, fnnam);
	} else {
		errtext2 = mystrdup((*yytext == '\'') ? yytext+1 : yytext, fnnam);
	}
	logreturn(PIKTLOG, LOG_ERR, ERRID_ALT_PARSE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s%s file, at or near line %d, %s: %s%s%s", altcurr, isalert ? ".alt" : nullstring, lineno, message, errtext1, (*errtext1 && *errtext2 && strcmp(errtext1,errtext2)) ? " or " : nullstring, strcmp(errtext1,errtext2) ? errtext2 : nullstring);
	BEGIN(INITIAL);
	bailout = TRUE;
	return;
}

/* feed command string to the input stream
 */
signed int
altinput (void)
{
	register char c;

	if (! *ccp) {
		return(EOF);
	}
	c = *ccp;
	ccp++;
	return(c);
}
