/*
   msg.c -- PIKT alert message routines

   $Header: /var/cvs/PIKT/pikt/src/piktd/msg.c,v 1.16 2006/11/04 09:59:36 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "../lib/common.h"

#include "piktd.h"

int newmailalert, newmailalarm, newprintalert, newprintalarm;

msgq *mailhdrmsg, *printhdrmsg;
msgq *mailalmmsg[8], *printalmmsg[8];
msgq *mailhdrmsgcurr, *printhdrmsgcurr;
msgq *mailalmmsgcurr[8], *printalmmsgcurr[8];

void regmsg(msgq **, msgq **, char *, int);

/* register alert header lines in the appropriate message queue
 */
void
altheader (msgq **msgqueue, msgq **msgqueuecurr, int type)
{
	struct tm *tm;
	time_t tnow;
	char altline[MAXLINE+1];
	char fmt[8];

	sprintf(altline, "\n%42s\n", "PIKT ALERT");
	regmsg(msgqueue, msgqueuecurr, altline, type);
	tnow = time((time_t*)0);
	tm = localtime(&tnow);
	sprintf(altline, "%50s", asctime(tm));
	regmsg(msgqueue, msgqueuecurr, altline, type);
	sprintf(fmt, "%%%ds\n", 37+(int)(strlen(hostname)/2));
	sprintf(altline, fmt, hostname);
	regmsg(msgqueue, msgqueuecurr, altline, type);
	return;
}

/* register alarm header lines in the appropriate message queue
 */
void
almheader (msgq **msgqueue, msgq **msgqueuecurr, int type)
{
	char almlin[MAXLINE+1];

	switch (LEVEL) {
		case LOG_EMERG:
			sprintf(almlin, "\n%s:\n", "EMERGENCY");
			break;
		case LOG_ALERT:
			sprintf(almlin, "\n%s:\n", "URGENT");
			break;
		case LOG_CRIT:
			sprintf(almlin, "\n%s:\n", "CRITICAL");
			break;
		case LOG_ERR:
			sprintf(almlin, "\n%s:\n", "ERROR");
			break;
		default:
			sprintf(almlin, "\n%s:\n", leveltypes[LEVEL]);
			break;
	}
	regmsg(msgqueue, msgqueuecurr, almlin, type);
	sprintf(almlin, "    %s\n", ALMNAM);
	regmsg(msgqueue, msgqueuecurr, almlin, type);
	requote(TASK);
	sprintf(almlin, "        %s\n\n", stripesc(TASK));
	regmsg(msgqueue, msgqueuecurr, almlin, type);
	return;
}

/* format the current alert message string; register the string in the
   appropriate message queue
 */
void
almline (msgq **msgqueue, msgq **msgqueuecurr, char *str, int type)
{
	char almlin[MAXBUF+1];
	char *s, *lf;

	s = str;
	if ((*s == '\n') && ! *(s+1))	/* str is lf only */
		s++;
	else
		while (*s && (lf = strchr(s, '\n'))) {
			*lf = '\0';
			sprintf(almlin, "%s%s\n", ! interactive ? "        " : nullstring, s);
			regmsg(msgqueue, msgqueuecurr, almlin, type);
			s = lf+1;
		}
	sprintf(almlin, "%s%s\n", ! interactive ? "        " : nullstring, s);
	regmsg(msgqueue, msgqueuecurr, almlin, type);
	return;
}

/* queue the given message string to the indicated queue
 */
void
regmsg (msgq **msgqueue, msgq **msgqueuecurr, char *line, int type)
{
	static char fnnam[] = "regmsg";
	msgq *nmp;

	if (! interactive && (type != OUTPUT)) {
		nmp = (msgq *) mymalloc(sizeof(msgq), fnnam);
		nmp->next = (msgq *) NULL;
		nmp->line = mystrdup(line, fnnam);
		if (! *msgqueue) {
			*msgqueue = nmp;
		} else {
			(*msgqueuecurr)->next = nmp;
		}
		*msgqueuecurr = nmp;
	} else {
		printf("%s", line);
		/* fflush(stdout); */
	}
	return;
}

/* dequeue the message lines to either an e-mail or print process
 */
/*
void
mysendmsg (msgq *messages, char *cmd)
{
	static char fnnam[] = "mysendmsg";
	msgq *mp;
	FILE *altfile;

	if (messages) {
		if (cmd) {
			if ((altfile = popen(cmd, "w")) == NULL)
				logexit(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "for alert %s, popen() error for %s", altcurr, cmd);
			foreach (mp, messages)
				fprintf(altfile, "%s", mp->line);
			pclose(altfile);
		} else
			logreturn(PIKTLOG, LOG_ERR, ERRID_MSG_CMD_NULL, __FILE__, __LINE__, fnnam, FALSE, "for alert %s, messages queued, but missing cmd", altcurr);
	}
	return;
}
*/

/* dequeue/garbage collect messages from the indicated queue
 */
void
freemsg (msgq **msgqueue)
{
	static msgq *mp, *tmp;

	mp = *msgqueue;
	while (mp) {
		tmp = mp;
		mp = mp->next;
		FREE(tmp->line);
		free(tmp);
	}
	*msgqueue = (msgq *) NULL;
	return;
}
