%{

/*
   timing.y -- PIKT piktd timing parser

   $Header: /var/cvs/PIKT/pikt/src/piktd/timing.y,v 1.21 2006/11/04 19:25:05 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktd.h"

char _timingparse_[] = "timingparse";

#define ISMON	(cmax == 12)	/* to identify mon field */
#define ISDOM	(cmax == 31)	/* to identify dom field */
#define ISDOW	(cmax ==  6)	/* to identify dow field */
#define ISHR	(cmax == 23)	/* to identify hr  field */
#define ISMIN	(cmax == 59)	/* to identify min field */
#define ISFTP	(etp == NONE)	/* to identify first time period */

int i;

%}

%union {
	int ival;
}

%token <ival> _NUMBER

%%

start:		'*' {
			for (i=clo; i<=chi; i++)
		  		cb |= (1UL << (i % 32));
			if (ISMON)
				mai = mop;
			else if (! ISDOW)
				mai = minval(mai, mop);
		}
		| '*' '/' _NUMBER {
			for (i=clo; i<=chi; i+=$3)
		  		cb |= (1UL << (i % 32));
			if (ISMON)
				mai = $3*mop;
			else
				mai = minval(mai, $3*mop);
		}
		| timingvals
		;

timingvals:	timingval
		| timingvals ',' timingval
		;

timingval:	_NUMBER {
			if (($1 >= cmin) && ($1 <= cmax)) {
				if (($1 >= clo) && ($1 <= chi))
					cb |= (1UL << ($1 % 32)) ;
			} else {
				logexit(PIKTLOG, LOG_ERR, ERRID_TIMING_VAL_INVALID, __FILE__, __LINE__, _timingparse_, FALSE, "in piktd.conf, bad timingval: %d", $1);
			}
			if (ISFTP) {
				etp = $1;
				if (ISMON || ISDOW)
					mai = map;
				else
					; /* do nothing! */
			} else {
				if ($1 <= ptp) {
					logexit(PIKTLOG, LOG_ERR, ERRID_TIMING_VAL_INVALID, __FILE__, __LINE__, _timingparse_, FALSE, "in piktd.conf, time periods must be in ascending order: %d,%d", ptp, $1);
				} else {
					mai = minval(mai, (($1-ptp)*mop));
					mai = minval(mai, ((ttp+etp-$1)*mop));
				}
			}
			ptp = $1;
		}
		| _NUMBER '-' _NUMBER {
			if ($1 >= $3) {
				logexit(PIKTLOG, LOG_ERR, ERRID_TIMING_VAL_INVALID, __FILE__, __LINE__, _timingparse_, FALSE, "in piktd.conf, bad timingval: %d-%d", $1, $3);
			}
			if (($1 >= cmin) && ($1 <= cmax)) {
				for (i=$1; i<=$3; i++)
					if ((i >= clo) && (i <= chi))
						cb |= (1UL << (i % 32));
			} else {
				logexit(PIKTLOG, LOG_ERR, ERRID_TIMING_VAL_INVALID, __FILE__, __LINE__, _timingparse_, FALSE, "in piktd.conf, bad timingval: %d-%d", $1, $3);
			}
			if (ISFTP) {
				etp = $1;
			} else {
				if ($1 <= ptp) {
					logexit(PIKTLOG, LOG_ERR, ERRID_TIMING_PERIODS_NOT_ASCENDING, __FILE__, __LINE__, _timingparse_, FALSE, "in piktd.conf, time periods must be in ascending order: %d,%d", ptp, $1);
				}
			}
			if (ISMON)
				mai = mop;
			else
				mai = minval(mai, mop);
			ptp = $3;
		}
		| _NUMBER '-' _NUMBER '/' _NUMBER {
			if (($1 >= $3) || ($5 > ($3 - $1))) {
				logexit(PIKTLOG, LOG_ERR, ERRID_TIMING_VAL_INVALID, __FILE__, __LINE__, _timingparse_, FALSE, "in piktd.conf, bad timingval: %d-%d/%d", $1, $3, $5);
			}
			if (($1 >= cmin) && ($1 <= cmax)) {
				for (i=$1; i<=$3; i+=$5)
					if ((i >= clo) && (i <= chi))
						cb |= (1UL << (i % 32));
			} else {
				logexit(PIKTLOG, LOG_ERR, ERRID_TIMING_VAL_INVALID, __FILE__, __LINE__, _timingparse_, FALSE, "in piktd.conf, bad timingval: %d-%d/%d", $1, $3, $5);
			}
			if (ISFTP) {
				etp = $1;
				if (ISMON)
					mai = $5*mop;
				else
					mai = minval(mai, $5*mop);
			} else {
				if ($1 <= ptp) {
					logexit(PIKTLOG, LOG_ERR, ERRID_TIMING_PERIODS_NOT_ASCENDING, __FILE__, __LINE__, _timingparse_, FALSE, "in piktd.conf, time periods must be in ascending order: %d,%d", ptp, $1);
				} else {
					mai = minval(mai, $5*mop);
					mai = minval(mai, (($1-ptp)*mop));
				}
			}
			ptp = $3;
			mai = minval(mai, ((ttp+etp-$5*(int)($3/$5))*mop));
		}
		| _NUMBER '%' {
			if (($1 >= 1) && ($1 <= 100)) {
				tp = (float)$1/(float)100;
			} else {
				logexit(PIKTLOG, LOG_ERR, ERRID_TIMING_VAL_INVALID, __FILE__, __LINE__, _timingparse_, FALSE, "in piktd.conf, bad timingval: %d%%", $1);
			}
			if (ISMON)
				mai = mop;
			else if (! ISDOW)
				mai = minval(mai, mop);
		}
		;

%%
